@@ -1,15 +1,19 @@
 """This module is the Python part of the CAD Viewer widget"""
 
+import base64
 import json
 from textwrap import dedent
-import ipywidgets as widgets
+
 import numpy as np
 
-from traitlets import Unicode, Dict, Tuple, Integer, Float, Any, Bool, observe
+import ipywidgets as widgets
+from ipywidgets.embed import embed_minimal_html, dependency_state
+
+from traitlets import Unicode, Dict, List, Tuple, Integer, Float, Any, Bool, observe
 from IPython.display import HTML, update_display
 from pyparsing import ParseException
 
-from .utils import serializer, get_parser
+from .utils import get_parser, to_json
 from .boundingbox import combined_bb, normalize
 
 
@@ -107,8 +111,8 @@ class CadViewerWidget(widgets.Output):  # pylint: disable-msg=too-many-instance-
     _model_name = Unicode("CadViewerModel").tag(sync=True)
     _view_module = Unicode("cad-viewer-widget").tag(sync=True)
     _model_module = Unicode("cad-viewer-widget").tag(sync=True)
-    _view_module_version = Unicode("0.10.2").tag(sync=True)
-    _model_module_version = Unicode("0.10.2").tag(sync=True)
+    _view_module_version = Unicode("1.1.1").tag(sync=True)
+    _model_module_version = Unicode("1.1.1").tag(sync=True)
 
     #
     # Display traits
@@ -124,46 +128,46 @@ class CadViewerWidget(widgets.Output):  # pylint: disable-msg=too-many-instance-
     "unicode string: Width of the canvas element"
 
     height = Integer(allow_none=True).tag(sync=True)
-    "int: Heigth of the canvas element"
+    "int: Height of the canvas element"
 
     tree_width = Integer(allow_none=True).tag(sync=True)
-    "int: Width of the navigatoin tree element"
+    "int: Width of the navigation tree element"
 
     theme = Unicode(allow_none=True).tag(sync=True)
     "unicode string: UI theme, can be 'dark' or 'light' (default)"
 
-    pinning = Bool(allow_none=True).tag(sync=True)
+    pinning = Bool(allow_none=True, default_value=None).tag(sync=True)
     "bool: Whether to show the pin a png button or not"
 
     #
     # Viewer traits
     #
 
-    shapes = Unicode(allow_none=True).tag(sync=True)
+    shapes = Dict(allow_none=True).tag(sync=True, to_json=to_json)
     "unicode: Serialized nested tessellated shapes"
 
     states = Dict(Tuple(Integer(), Integer()), allow_none=True).tag(sync=True)
     "dict: State of the nested cad objects, key = object path, value = 2-dim tuple of 0/1 (hidden/visible) for object and edges"
 
-    tracks = Unicode(allow_none=True).tag(sync=True)
+    tracks = List(allow_none=True).tag(sync=True)
     "unicode: Serialized list of animation track arrays, see [AnimationTrack.to_array](/widget.html#cad_viewer_widget.widget.AnimationTrack.to_array)"
 
-    timeit = Bool(allow_none=True).tag(sync=True)
+    timeit = Bool(allow_none=True, default_value=None).tag(sync=True)
     "bool: Whether to output timing info to the browser console (True) or not (False)"
 
-    tools = Bool(allow_none=True).tag(sync=True)
+    tools = Bool(allow_none=True, default_value=None).tag(sync=True)
     "bool: Whether to show CAD tools (True) or not (False)"
 
-    ortho = Bool(allow_none=True).tag(sync=True)
+    ortho = Bool(allow_none=True, default_value=None).tag(sync=True)
     "bool: Whether to use orthographic view (True) or perspective view (False)"
 
     control = Unicode().tag(sync=True)
     "unicode: Whether to use trackball controls ('trackball') or orbit controls ('orbit')"
 
-    axes = Bool(allow_none=True).tag(sync=True)
+    axes = Bool(allow_none=True, default_value=None).tag(sync=True)
     "bool: Whether to show coordinate axes (True) or not (False)"
 
-    axes0 = Bool(allow_none=True).tag(sync=True)
+    axes0 = Bool(allow_none=True, default_value=None).tag(sync=True)
     "bool: Whether to center coordinate axes at the origin [0,0,0] (True) or at the CAD object center (False)"
 
     grid = Tuple(Bool(), Bool(), Bool(), allow_none=True).tag(sync=True)
@@ -172,10 +176,10 @@ class CadViewerWidget(widgets.Output):  # pylint: disable-msg=too-many-instance-
     ticks = Integer(allow_none=True).tag(sync=True)
     "integer: Hint for the number of ticks for the grids (will be adjusted for nice intervals)"
 
-    transparent = Bool(allow_none=True).tag(sync=True)
-    "bool: Whether to show the CAD objetcs transparently (True) or not (False)"
+    transparent = Bool(allow_none=True, default_value=None).tag(sync=True)
+    "bool: Whether to show the CAD objects transparently (True) or not (False)"
 
-    black_edges = Bool(allow_none=True).tag(sync=True)
+    black_edges = Bool(allow_none=True, default_value=None).tag(sync=True)
     "bool: Whether to shows the edges in black (True) or not(False)"
 
     normal_len = Float(allow_none=True).tag(sync=True)
@@ -198,33 +202,33 @@ class CadViewerWidget(widgets.Output):  # pylint: disable-msg=too-many-instance-
     #
 
     tab = Unicode(allow_none=True).tag(sync=True)
-    "unicode: Whther to show the navigation tree ('tree') or the clipping UI ('clip')"
+    "unicode: Whether to show the navigation tree ('tree') or the clipping UI ('clip')"
 
-    clip_intersection = Bool(allow_none=True).tag(sync=True)
+    clip_intersection = Bool(allow_none=True, default_value=None).tag(sync=True)
     "bool: Whether to use intersection clipping (True) or not (False)"
 
-    clip_planes = Bool(allow_none=True).tag(sync=True)
+    clip_planes = Bool(allow_none=True, default_value=None).tag(sync=True)
     "bool: Whether to show colored clipping planes (True) or not (False)"
 
-    clip_normal_0 = Tuple(Float(), Float(), Float(), allow_none=True).tag(sync=True)
+    clip_normal_0 = Tuple(Float(), Float(), Float(), allow_none=True, default_value=None).tag(sync=True)
     "tuple: Normal of clipping plane 1 as a 3-dim tuple of float (x,y,z)"
 
-    clip_normal_1 = Tuple(Float(), Float(), Float(), allow_none=True).tag(sync=True)
+    clip_normal_1 = Tuple(Float(), Float(), Float(), allow_none=True, default_value=None).tag(sync=True)
     "tuple: Normal of clipping plane 2 as a 3-dim tuple of float (x,y,z)"
 
-    clip_normal_2 = Tuple(Float(), Float(), Float(), allow_none=True).tag(sync=True)
+    clip_normal_2 = Tuple(Float(), Float(), Float(), allow_none=True, default_value=None).tag(sync=True)
     "tuple: Normal of clipping plane 3 as a 3-dim tuple of float (x,y,z)"
 
-    clip_slider_0 = Float(allow_none=True).tag(sync=True)
+    clip_slider_0 = Float(allow_none=True, default_value=None).tag(sync=True)
     "float: Slider value of clipping plane 1"
 
-    clip_slider_1 = Float(allow_none=True).tag(sync=True)
+    clip_slider_1 = Float(allow_none=True, default_value=None).tag(sync=True)
     "float: Slider value of clipping plane 2"
 
-    clip_slider_2 = Float(allow_none=True).tag(sync=True)
+    clip_slider_2 = Float(allow_none=True, default_value=None).tag(sync=True)
     "float: Slider value of clipping plane 3"
 
-    reset_camera = Bool(allow_none=True).tag(sync=True)
+    reset_camera = Bool(allow_none=True, default_value=None).tag(sync=True)
     "bool: Whether to reset camera (True) or not (False)"
 
     position = Tuple(Float(), Float(), Float(), allow_none=True).tag(sync=True)
@@ -233,6 +237,9 @@ class CadViewerWidget(widgets.Output):  # pylint: disable-msg=too-many-instance-
     quaternion = Tuple(Float(), Float(), Float(), Float(), allow_none=True).tag(sync=True)
     "tuple: Rotation of the camera as 4-dim quaternion (x,y,z,w)"
 
+    target = Tuple(Float(), Float(), Float(), allow_none=True).tag(sync=True)
+    "tuple: Camera target to look at as 3-dim tuple (x,y,z)"
+
     zoom = Float(allow_none=True).tag(sync=True)
     "float: Zoom value of the camera"
 
@@ -243,7 +250,10 @@ class CadViewerWidget(widgets.Output):  # pylint: disable-msg=too-many-instance-
     "tuple: Initial rotation of the camera as 4-dim quaternion (x,y,z,w)"
 
     zoom0 = Float(allow_none=True).tag(sync=True)
-    "float: Inital zoom value of the camera"
+    "float: Initial zoom value of the camera"
+
+    target0 = Tuple(Float(), Float(), Float(), allow_none=True).tag(sync=True)
+    "tuple: Initial camera target to look at as 3-dim tuple (x,y,z)"
 
     zoom_speed = Float(allow_none=True).tag(sync=True)
     "float: Speed of zooming with the mouse"
@@ -267,23 +277,20 @@ class CadViewerWidget(widgets.Output):  # pylint: disable-msg=too-many-instance-
     lastPick = Dict(key_trait=Unicode(), value_trait=Any(), allow_none=True, read_only=True).tag(sync=True)
     "dict: Describes the last picked element of the CAD view"
 
-    target = Tuple(Float(), Float(), Float(), allow_none=True, read_only=True).tag(sync=True)
-    "tuple: Camera target as a 3-dim tuple of float (x,y,z)"
-
-    result = Unicode(allow_none=True, read_only=True).tag(sync=True)
-    "unicode string: JSON serialiued result from Javascript"
+    result = Unicode(allow_none=True).tag(sync=True)
+    "unicode string: JSON serialized result from Javascript"
 
     #
     # Internal traitlets
     #
 
-    disposed = Bool(default=False, allow_none=True).tag(sync=True)
+    disposed = Bool(default=False, allow_none=True, default_value=None).tag(sync=True)
     "unicode string: Whether the Javascript viewer is disposed"
 
-    initialize = Bool(allow_none=True).tag(sync=True)
-    "bool: internally used to control initialisation of view. Do not use!"
+    initialize = Bool(allow_none=True, default_value=None).tag(sync=True)
+    "bool: internally used to control initialization of view. Do not use!"
 
-    js_debug = Bool(allow_none=True).tag(sync=True)
+    js_debug = Bool(allow_none=True, default_value=None).tag(sync=True)
     "bool: Whether to show infos in the browser console (True) or not (False)"
 
     image_id = Unicode(allow_none=True).tag(sync=True)
@@ -291,9 +298,20 @@ class CadViewerWidget(widgets.Output):  # pylint: disable-msg=too-many-instance-
 
     @observe("result")
     def func(self, change):
-        data = json.loads(change["new"])
-        html = f"""<img src="{data['src']}" width="{data['width']}px" height="{data['height']}px"/>"""
-        update_display(HTML(html), display_id=data["display_id"])
+        if change["new"] is not None:
+            data = json.loads(change["new"])
+
+            if data.get("display_id") is not None:
+                html = f"""<img src="{data['src']}" width="{data['width']}px" height="{data['height']}px"/>"""
+                update_display(HTML(html), display_id=data["display_id"])
+            else:
+                if self.test_func is not None and callable(self.test_func):
+                    self.test_func(base64.b64decode(data["src"].split(",")[1]))  # pylint: disable=not-callable
+                else:
+                    with open(data["filename"], "wb") as fd:
+                        fd.write(base64.b64decode(data["src"].split(",")[1]))
+
+            self.result = None
 
 
 class CadViewer:
@@ -305,9 +323,9 @@ class CadViewer:
     cad_width : int, default: 800
         Width of the canvas element
     height : int, default: 600
-        Heigth of the canvas element
+        Height of the canvas element
     tree_width : int, default: 240
-        Width of the navigatoin tree element
+        Width of the navigation tree element
     theme : string, default: 'light'
         UI theme, can be 'dark' or 'light' (default)
     tools : bool, default: True
@@ -347,12 +365,11 @@ class CadViewer:
             title=title,
             anchor=anchor,
         )
+        self.widget.test_func = None
         self.msg_id = 0
         self.parser = get_parser()
 
-        self.widget.position0 = None
-        self.widget.quaternion0 = None
-        self.widget.zoom0 = None
+        self.empty = True
 
         self.tracks = []
 
@@ -392,6 +409,7 @@ class CadViewer:
         black_edges=False,
         position=None,
         quaternion=None,
+        target=None,
         zoom=1.0,
         reset_camera=True,
         zoom_speed=1.0,
@@ -399,7 +417,6 @@ class CadViewer:
         rotate_speed=1.0,
         timeit=False,
         js_debug=False,
-        # bb_factor=1.0,
     ):
         """
         Adding shapes to the CAD view
@@ -427,7 +444,7 @@ class CadViewer:
         ticks : int, default 10
             Hint for the number of ticks for the grids (will be adjusted for nice intervals)
         transparent : bool, default False
-            Whether to show the CAD objetcs transparently (True) or not (False)
+            Whether to show the CAD objects transparently (True) or not (False)
         black_edges : bool, default False
             Whether to shows the edges in black (True) or not(False)
         normal_Len : int, default 0
@@ -444,6 +461,8 @@ class CadViewer:
             Position of the camera as a 3-dim tuple of float (x,y,z)
         quaternion : 4-dim list of float, default None
             Rotation of the camera as 4-dim quaternion (x,y,z,w)
+        target :  3-dim list of float, default None
+            Camera target to look at, default is the center of the object's bounding box
         zoom : float, default None
             Zoom value of the camera
         reset_camera : bool, default True
@@ -460,7 +479,7 @@ class CadViewer:
         Examples
         --------
 
-        A simple cube with edge len of 1 is tessellated like the `shape` eloment of the first (and only) element of
+        A simple cube with edge len of 1 is tessellated like the `shape` element of the first (and only) element of
         the `parts` list:
 
         ```
@@ -637,18 +656,22 @@ class CadViewer:
             grid = [False, False, False]
 
         # If one changes the control type, override reset_camera with "True"
-        if self.widget.control != control:
+        if self.empty or self.widget.control != control:
             reset_camera = True
+            self.empty = False
+
             # Don't show warning on first call
             if self.widget.control != "":
                 print("Camera control changed, so camera was resetted")
 
         if reset_camera:
+            bb = combined_bb(shapes)
+
             if position is None:
-                bb = combined_bb(shapes)
                 position = (
                     normalize(np.array((1, 1, 1))) * 5.5 * bb.max_dist_from_center() + np.array(bb.center)
                 ).tolist()
+
             if quaternion is None and control == "trackball":
                 quaternion = (
                     0.1759198966061612,
@@ -656,17 +679,23 @@ class CadViewer:
                     0.8204732385702833,
                     0.33985114297998736,
                 )
+
+            if target is None:
+                target = bb.center
+
             if zoom is None:
                 zoom = 1.0
         else:
-            position = (*self.widget.position,)
-            quaternion = (*self.widget.quaternion,)
+            copy = lambda v: None if v is None else (*v,)
+            position = copy(self.widget.position)
+            quaternion = copy(self.widget.quaternion)
+            target = copy(self.widget.target)
             zoom = self.widget.zoom
 
         self.widget.initialize = True
 
         with self.widget.hold_trait_notifications():
-            self.widget.shapes = json.dumps(shapes, default=serializer)
+            self.widget.shapes = shapes
             self.widget.states = states
 
             self.widget.default_edge_color = default_edge_color
@@ -687,7 +716,15 @@ class CadViewer:
             self.widget.reset_camera = reset_camera
             self.widget.position = position
             self.widget.quaternion = quaternion
+            self.widget.target = target
             self.widget.zoom = zoom
+            # If reset_camera, position0, quaternion0 and zoom0 need to be set
+            if reset_camera:
+                self.widget.position0 = (*position,)
+                if control == "trackball":
+                    self.widget.quaternion0 = (*quaternion,)
+                self.widget.target0 = (*target,)
+                self.widget.zoom0 = zoom
             self.widget.zoom_speed = zoom_speed
             self.widget.pan_speed = pan_speed
             self.widget.rotate_speed = rotate_speed
@@ -696,7 +733,7 @@ class CadViewer:
             self.add_tracks(tracks)
 
         self.widget.initialize = False
-        self.update_camera_location()
+        # self.update_camera_location()
 
     def update_states(self, states):
         """Set navigation tree states for a CAD view"""
@@ -704,7 +741,7 @@ class CadViewer:
         self.widget.state_updates = states
 
     def update_camera_location(self):
-        """Sync position, quaternion and zoom of camer to Python"""
+        """Sync position, quaternion and zoom of camera to Python"""
         self.execute("updateCamera", [])
 
     def close(self):
@@ -1049,7 +1086,6 @@ class CadViewer:
         see [CadViewerWidget.zoom](./widget.html#cad_viewer_widget.widget.CadViewerWidget.zoom)
         """
 
-        self.update_camera_location()
         return self.widget.zoom
 
     @zoom.setter
@@ -1063,7 +1099,6 @@ class CadViewer:
         see [CadViewerWidget.position](./widget.html#cad_viewer_widget.widget.CadViewerWidget.position)
         """
 
-        self.update_camera_location()
         return self.widget.position
 
     @position.setter
@@ -1077,15 +1112,32 @@ class CadViewer:
         see [CadViewerWidget.quaternion](./widget.html#cad_viewer_widget.widget.CadViewerWidget.quaternion)
         """
 
-        self.update_camera_location()
         if self.widget.control == "orbit":
+            print("quaternion controlled internally for control=='orbit'")
             return None
         else:
             return self.widget.quaternion
 
     @quaternion.setter
     def quaternion(self, value):
-        self.widget.quaternion = value
+        if self.widget.control == "orbit":
+            print("quaternion controlled internally for control=='orbit'")
+        else:
+            self.widget.quaternion = value
+
+    @property
+    def target(self):
+        """
+        Get or set the CadViewerWidget traitlet `target`
+        see [CadViewerWidget.position](./widget.html#cad_viewer_widget.widget.CadViewerWidget.target)
+        """
+
+        # self.update_camera_location()
+        return self.widget.target
+
+    @target.setter
+    def target(self, value):
+        self.widget.target = value
 
     @property
     def last_pick(self):
@@ -1105,6 +1157,18 @@ class CadViewer:
 
         return self.widget.control
 
+    @property
+    def pinning(self):
+        """
+        Get or set the CadViewerWidget traitlet `pinning`
+        see [CadViewerWidget.pinning](./widget.html#cad_viewer_widget.widget.CadViewerWidget.pinning)
+        """
+        return self.widget.pinning
+
+    @pinning.setter
+    def pinning(self, flag):
+        self.widget.pinning = flag
+
     #
     # Animation handling
     #
@@ -1115,7 +1179,40 @@ class CadViewer:
         """
 
         self.tracks = []
-        self.widget.tracks = ""
+        self.widget.tracks = []
+
+    def _check_track(self, track):
+        paths = self.widget.states.keys()
+        if not any([(f"{path}/").startswith(f"{track.path}/") for path in paths]):
+            raise ValueError(f"{track.path} is not a valid subpath of any of {list(paths)}")
+
+        actions = ["t", "tx", "ty", "tz", "q", "rx", "ry", "rz"]
+        if not track.action in actions:
+            raise ValueError(f"{track.action} is not a valid action {list(actions)}")
+
+        if len(track.times) != len(track.values):
+            raise ValueError(f"Track times and values need to have same length")
+
+        if not all([isinstance(t, (int, float)) for t in track.times]):
+            raise ValueError(f"Time values need to be int or float")
+
+        if track.action in ["tx", "ty", "tz", "rx", "ry", "rz"]:
+            if not all([isinstance(t, (int, float)) for t in track.values]):
+                raise ValueError(f"Value values need to be int or float for action '{track.action}'")
+
+        if track.action in ["t", "q"]:
+            size = 3 if track.action == "t" else 4
+            if not all(
+                [
+                    isinstance(v, (list, tuple)) and (len(v) == size) and all([isinstance(x, (int, float)) for x in v])
+                    for v in track.values
+                ]
+            ):
+                raise ValueError(
+                    f"Value values need to be {size} dim lists of int or float for action '{track.action}'"
+                )
+
+        return track
 
     def add_track(self, track):
         """
@@ -1127,7 +1224,7 @@ class CadViewer:
             Animation track, see [AnimationTrack](/widget.html#cad_viewer_widget.widget.AnimationTrack)
         """
 
-        self.tracks.append(track)
+        self.tracks.append(self._check_track(track))
 
     def add_tracks(self, tracks):
         """
@@ -1138,8 +1235,10 @@ class CadViewer:
         tracks: list of AnimationTrack
             List of Animation tracks, see [AnimationTrack](/widget.html#cad_viewer_widget.widget.AnimationTrack)
         """
-
-        self.tracks = [] if tracks is None else [track for track in tracks]  # enforce a new array
+        checked_tracks = (
+            [] if tracks is None else [self._check_track(track) for track in tracks]
+        )  # enforce a new array
+        self.tracks = checked_tracks
 
     def animate(self, speed=1):
         """
@@ -1151,7 +1250,7 @@ class CadViewer:
             Animation speed, will be forwarded via `animation_speed` traitlet
         """
 
-        self.widget.tracks = json.dumps([track.to_array() for track in self.tracks])
+        self.widget.tracks = [track.to_array() for track in self.tracks]
         self.widget.animation_speed = speed
         self.execute("animate")
         # self.play()
@@ -1182,7 +1281,14 @@ class CadViewer:
         Pin CAD View as PNG
         """
 
-        self.execute("viewer.pinAsPng", None)
+        self.execute("pinAsPng", None)
+
+    def export_png(self, filename):
+        """
+        Save CAD View as PNG
+        """
+
+        self.execute("saveAsPng", filename)
 
     #
     # Tab handling
@@ -1206,6 +1312,19 @@ class CadViewer:
     # Rotations
     #
 
+    def set_camera(self, direction):
+        """
+        Set camera to one of the predefined locations
+
+        Parameters
+        ----------
+        direction : string
+            one of ["iso", "top", "bottom", "left", "right", "front", "rear"]
+        """
+
+        self.execute("viewer.camera.presetCamera", [direction])
+        self.execute("viewer.update", [])
+
     def rotate_x(self, angle):
         """
         Rotate CAD obj around x-axis - trackball controls only
@@ -1276,6 +1395,21 @@ class CadViewer:
             raise NameError("rotateLeft only works for orbit control")
         self.execute("viewer.controls.rotateLeft", (angle,))
 
+    #
+    # Exports
+    #
+
+    def export_html(self, filename="cadquery.html", title="CadQuery"):
+        if self.widget.title != None:
+            raise RuntimeError("Export_html does not work with sidecar. Show the object again in a cell viewer")
+
+        pinning = self.pinning
+        self.pinning = False
+
+        embed_minimal_html(filename, title=title, views=[self.widget], state=dependency_state(self.widget))
+
+        self.pinning = pinning
+
     #
     # Custom message handling
     #
@@ -1327,7 +1461,7 @@ class CadViewer:
                 pinning:            {self.widget.pinning}
 
                             SHAPES
-                shapes:             {self.widget.shapes if shapes else (self.widget.shapes[:200] + " ...")}
+                shapes:             {self.widget.shapes if shapes else "... (set shapes=True)"}
                 states:             {self.widget.states}
                 tracks:             {self.widget.tracks}
                             
@@ -1361,11 +1495,12 @@ class CadViewer:
                 reset_camera:       {self.widget.reset_camera}
                 position:           {self.widget.position}
                 quaternion:         {self.widget.quaternion}
+                target:             {self.widget.target}
                 zoom:               {self.widget.zoom}
                 position0:          {self.widget.position0}
                 quaternion0:        {self.widget.quaternion0}
+                target0:            {self.widget.target0}
                 zoom0:              {self.widget.zoom0}
-                target:             {self.widget.target}
                 zoom_speed:         {self.widget.zoom_speed}
                 pan_speed:          {self.widget.pan_speed}
                 rotate_speed:       {self.widget.rotate_speed}
