@@ -70,10 +70,12 @@ export class CadViewerModel extends DOMWidgetModel {
 
       position: null,
       quaternion: null,
+      target: null,
       zoom: null,
 
       position0: null,
       quaternion0: null,
+      target0: null,
       zoom0: null,
 
       zoom_speed: null,
@@ -84,7 +86,6 @@ export class CadViewerModel extends DOMWidgetModel {
       // Read only traitlets
 
       lastPick: null,
-      target: null,
 
       initialize: null,
       image_id: null,
@@ -100,6 +101,15 @@ export class CadViewerModel extends DOMWidgetModel {
 export class CadViewerView extends DOMWidgetView {
   initialize(...args) {
     super.initialize(...args);
+    this.position0 = null;
+    this.quaternion0 = null;
+    this.target0 = null;
+    this.zoom0 = null;
+    this.lastPosition = null;
+    this.lastQuaternion = null;
+    this.lastTarget = null;
+    this.lastZoom = null;
+    this.empty = true;
   }
 
   debug(...args) {
@@ -115,6 +125,7 @@ export class CadViewerView extends DOMWidgetView {
       this.model.on("change:tracks", this.handle_change, this);
       this.model.on("change:position", this.handle_change, this);
       this.model.on("change:quaternion", this.handle_change, this);
+      this.model.on("change:target", this.handle_change, this);
       this.model.on("change:zoom", this.handle_change, this);
       this.model.on("change:axes", this.handle_change, this);
       this.model.on("change:grid", this.handle_change, this);
@@ -123,6 +134,7 @@ export class CadViewerView extends DOMWidgetView {
       this.model.on("change:transparent", this.handle_change, this);
       this.model.on("change:black_edges", this.handle_change, this);
       this.model.on("change:tools", this.handle_change, this);
+      this.model.on("change:pinning", this.handle_change, this);
       this.model.on("change:default_edge_color", this.handle_change, this);
       this.model.on("change:default_opacity", this.handle_change, this);
       this.model.on("change:ambient_intensity", this.handle_change, this);
@@ -144,12 +156,13 @@ export class CadViewerView extends DOMWidgetView {
       this.model.on("change:js_debug", this.handle_change, this);
       this.model.on("change:disposed", this.handle_change, this);
 
-      // this.model.on("change:bb_factor", this.handle_change, this);
-
       this.listenTo(this.model, "msg:custom", this.onCustomMessage.bind(this));
 
       this.shell = App.getShell();
 
+      // in case of embedding we need to state values later, since rendering resets them
+      this.backupClipping();
+
       this.init = false;
       this.disposed = false;
 
@@ -210,6 +223,7 @@ export class CadViewerView extends DOMWidgetView {
       rotateSpeed: this.model.get("rotate_speed"),
       position: this.model.get("position"),
       quaternion: this.model.get("quaternion"),
+      target: this.model.get("target"),
       zoom: this.model.get("zoom")
     };
 
@@ -260,7 +274,7 @@ export class CadViewerView extends DOMWidgetView {
       container,
       options,
       this.handleNotification.bind(this),
-      this.pinAsPng.bind(this),
+      this.exportPng.bind(this),
       true
     );
 
@@ -285,6 +299,18 @@ export class CadViewerView extends DOMWidgetView {
     this.init = this.model.get("initialize");
 
     if (this.init) {
+      // support rest initial position and  keeping camera location
+      if (!this.empty) {
+        this.position0 = this.model.get("position0");
+        this.quaternion0 = this.model.get("quaternion0");
+        this.zoom0 = this.model.get("zoom0");
+        this.target0 = this.model.get("target0");
+
+        this.lastPosition = this.viewer.getCameraPosition();
+        this.lastQuaternion = this.viewer.getCameraQuaternion();
+        this.lastZoom = this.viewer.getCameraZoom();
+        this.lastTarget = this.viewer.getCameraTarget();
+      }
       this.clear();
     } else {
       const states = this.model.get("states");
@@ -303,33 +329,70 @@ export class CadViewerView extends DOMWidgetView {
     return states2;
   }
 
+  backupClipping() {
+    this.clipSettings = {
+      tab: this.model.get("tab"),
+      clip_planes: this.model.get("clip_planes"),
+      clip_intersection: this.model.get("clip_intersection"),
+      clip_normal_0: this.model.get("clip_normal_0"),
+      clip_normal_1: this.model.get("clip_normal_1"),
+      clip_normal_2: this.model.get("clip_normal_2"),
+      clip_slider_0: this.model.get("clip_slider_0"),
+      clip_slider_1: this.model.get("clip_slider_1"),
+      clip_slider_2: this.model.get("clip_slider_2")
+    };
+  }
+
+  setClipping() {
+    if (this.clipSettings.tab != null) {
+      this.viewer.display.selectTabByName(this.clipSettings.tab);
+    }
+    if (this.clipSettings.clip_intersection != null) {
+      this.viewer.setClipIntersection(
+        this.clipSettings.clip_intersection,
+        false
+      );
+    }
+    if (this.clipSettings.clip_planes != null) {
+      this.viewer.setClipPlaneHelpers(this.clipSettings.clip_planes, false);
+    }
+    if (this.clipSettings.clip_normal_0 != null) {
+      this.viewer.setClipNormal(0, this.clipSettings.clip_normal_0, false);
+    }
+    if (this.clipSettings.clip_normal_1 != null) {
+      this.viewer.setClipNormal(1, this.clipSettings.clip_normal_1, false);
+    }
+    if (this.clipSettings.clip_normal_2 != null) {
+      this.viewer.setClipNormal(2, this.clipSettings.clip_normal_2, false);
+    }
+    if (this.clipSettings.clip_slider_0 != null) {
+      this.viewer.setClipSlider(0, this.clipSettings.clip_slider_0, false);
+    }
+    if (this.clipSettings.clip_slider_1 != null) {
+      this.viewer.setClipSlider(1, this.clipSettings.clip_slider_1, false);
+    }
+    if (this.clipSettings.clip_slider_2 != null) {
+      this.viewer.setClipSlider(2, this.clipSettings.clip_slider_2, false);
+    }
+  }
+
   addShapes() {
+    if (this.model.get("initialize") == null) {
+      return;
+    }
+
+    this.shapes = this.model.get("shapes");
+    decode(this.shapes);
+    this.states = this.clone_states();
+
     const timer = new Timer("addShapes", this.model.get("timeit"));
     this._debug = this.model.get("js_debug");
 
     const resetCamera = this.model.get("reset_camera");
-    var position0 = null;
-    var quaternion0 = null;
-    var zoom0 = null;
-    var position1 = null;
-    var quaternion1 = null;
-    var zoom1 = null;
+    var position = null;
+    var quaternion = null;
     var target = null;
-    
-    if (!resetCamera) {
-      // get both, old reset location and current location of camera
-      position0 = this.model.get("position0");
-      quaternion0 = this.model.get("quaternion0");
-      zoom0 = this.model.get("zoom0");
-      target = this.model.get("target");
-
-      position1 = this.viewer.getCameraPosition();
-      quaternion1 = this.viewer.getCameraQuaternion();
-      zoom1 = this.viewer.getCameraZoom();
-    }
-
-    this.shapes = decode(this.model.get("shapes"));
-    this.states = this.clone_states();
+    var zoom = null;
 
     this.tracks = [];
 
@@ -348,39 +411,43 @@ export class CadViewerView extends DOMWidgetView {
 
     timer.split("renderer");
 
-    var position, quaternion, zoom;
-
-    if (resetCamera) {
-      // after the first view store inital camera location
-      zoom = this.viewer.getCameraZoom();
+    if (this.empty || resetCamera) {
+      // store inital camera location
       position = this.viewer.getCameraPosition();
       quaternion = this.viewer.getCameraQuaternion();
+      target = this.viewer.getCameraTarget();
+      zoom = this.viewer.getCameraZoom();
 
-      this.model.set("zoom0", zoom);
-      this.model.set("position0", position);
-      this.model.set("quaternion0", quaternion);
+      this.empty = false;
     } else {
       this.viewer.setResetLocation(
-        target,
-        position0,
-        quaternion0,
-        zoom0
+        this.target0,
+        this.position0,
+        this.quaternion0,
+        this.zoom0
       );
 
-      position = [...position1];
-      quaternion = [...quaternion1];
-      zoom = zoom1;
-      this.viewer.setCameraPosition(position);
-      this.viewer.setCameraQuaternion(quaternion)
-      this.viewer.setCameraZoom(zoom1);
+      position = [...this.lastPosition];
+      quaternion = [...this.lastQuaternion];
+      target = [...this.lastTarget];
+      zoom = this.lastZoom;
+      this.viewer.setCameraLocationSettings(
+        this.lastPosition,
+        this.lastQuaternion,
+        this.lastTarget,
+        this.lastZoom
+      );
     }
 
-    this.model.set("zoom", zoom);
     this.model.set("position", position);
     this.model.set("quaternion", quaternion);
+    this.model.set("target", target);
+    this.model.set("zoom", zoom);
 
     this.model.save_changes();
 
+    this.setClipping();
+
     // add animation tracks if exist
     const tracks = this.model.get("tracks");
     if (tracks != "" && tracks != null) {
@@ -392,20 +459,21 @@ export class CadViewerView extends DOMWidgetView {
 
     return true;
   }
-  
+
   updateCamera() {
-      var zoom = this.viewer.getCameraZoom();
-      var position = this.viewer.getCameraPosition();
-      var quaternion = this.viewer.getCameraQuaternion();
-
-      this.model.set("zoom", zoom);
-      this.model.set("position", position);
-      this.model.set("quaternion", quaternion);
+    var zoom = this.viewer.getCameraZoom();
+    var position = this.viewer.getCameraPosition();
+    var quaternion = this.viewer.getCameraQuaternion();
+    var target = this.viewer.getCameraTarget();
 
+    this.model.set("zoom", zoom);
+    this.model.set("position", position);
+    this.model.set("quaternion", quaternion);
+    this.model.set("target", target);
   }
 
   addTracks(tracks) {
-    this.tracks = decode(tracks);
+    this.tracks = tracks;
     if (Array.isArray(this.tracks) && this.tracks.length > 0) {
       for (var track of this.tracks) {
         this.viewer.addAnimationTrack(...track);
@@ -468,6 +536,9 @@ export class CadViewerView extends DOMWidgetView {
       case "quaternion":
         setKey("getCameraQuaternion", "setCameraQuaternion", key);
         break;
+      case "target":
+        setKey("getCameraTarget", "setCameraTarget", key);
+        break;
       case "axes":
         setKey("getAxes", "setAxes", key);
         break;
@@ -489,6 +560,10 @@ export class CadViewerView extends DOMWidgetView {
       case "tools":
         setKey("getTools", "setTools", key);
         break;
+      case "pinning":
+        var flag = change.changed[key];
+        this.viewer.display.setPinning(flag);
+        break;
       case "default_edge_color":
         setKey("getEdgeColor", "setEdgeColor", key);
         break;
@@ -573,20 +648,42 @@ export class CadViewerView extends DOMWidgetView {
     }
   }
 
-  pinAsPng(image) {
-    this.model.set(
-      "result",
-      JSON.stringify({
-        display_id: this.model.get("image_id"),
-        src: image.src,
-        width: image.width,
-        height: image.height
-      })
-    );
-    this.model.save_changes();
-    // and remove itself
-    this.dispose();
-    App.removeCellViewer(this.container_id);
+  exportPng(image) {
+    if (this.png_filename == null) {
+      this.model.set(
+        "result",
+        JSON.stringify({
+          display_id: this.model.get("image_id"),
+          src: image.src,
+          width: image.width,
+          height: image.height
+        })
+      );
+      this.model.save_changes();
+
+      this.dispose();
+      App.removeCellViewer(this.container_id);
+    } else {
+      this.model.set(
+        "result",
+        JSON.stringify({
+          filename: this.png_filename,
+          src: image.src
+        })
+      );
+      this.model.save_changes();
+      this.png_filename = null;
+    }
+  }
+
+  saveAsPng(filename) {
+    this.png_filename = filename;
+    this.viewer.pinAsPng();
+  }
+
+  pinAsPng() {
+    this.png_filename = null;
+    this.viewer.pinAsPng();
   }
 
   onCustomMessage(msg, buffers) {
